"""
Main smartglass client

Common script that handles several subcommands
See `Commands`
"""
import sys
import logging
import argparse
import functools

from logging.handlers import RotatingFileHandler
from code import InteractiveConsole

from gevent import signal, backdoor

from xbox.scripts import TOKENS_FILE, CONSOLES_FILE, LOG_FMT, \
    LOG_LEVEL_DEBUG_INCL_PACKETS, VerboseFormatter, ExitCodes

from xbox.handlers import tui, gamepad_input, text_input, fallout4_relay
from xbox.auxiliary.manager import TitleManager

from xbox.webapi.authentication.manager import AuthenticationManager
from xbox.webapi.common.exceptions import AuthenticationException

from xbox.sg import manager
from xbox.sg.console import Console
from xbox.sg.enum import ConnectionState

# REST server imports
from gevent import pywsgi as rest_pywsgi
from xbox.rest.app import app as flask_app


LOGGER = logging.getLogger(__name__)

REST_DEFAULT_SERVER_PORT = 5557
REPL_DEFAULT_SERVER_PORT = 5558


class Commands(object):
    """
    Available commands for CLI
    """
    Discover = 'discover'
    PowerOn = 'poweron'
    PowerOff = 'poweroff'
    REPL = 'repl'
    REPLServer = 'replserver'
    FalloutRelay = 'forelay'
    GamepadInput = 'gamepadinput'
    TextInput = 'textinput'
    TUI = 'tui'
    RESTServer = 'rest'


def parse_arguments(args=None):
    """
    Parse arguments with argparse.ArgumentParser

    Returns:
        Namespace: Parsed arguments

    Raises:
        Exception: On generic failure
    """

    parser = argparse.ArgumentParser(description='Xbox SmartGlass client')

    """Common arguments for logging"""
    logging_args = argparse.ArgumentParser(add_help=False)
    logging_args.add_argument(
        '--logfile',
        help="Path for logfile")
    logging_args.add_argument(
        '-v', '--verbose', action='count', default=0,
        help='Set logging level\n'
             '(  -v: INFO,\n'
             ' -vv: DEBUG,\n'
             '-vvv: DEBUG_INCL_PACKETS)')

    """Common arguments for authenticated console connection"""
    xbl_token_args = argparse.ArgumentParser(add_help=False)
    xbl_token_args.add_argument(
        '--tokens', '-t', type=str, default=TOKENS_FILE,
        help='Tokenfile to load')
    xbl_token_args.add_argument(
        '--refresh', '-r', action='store_true',
        help="Refresh xbox live tokens in provided token file")

    """Common argument for console connection"""
    connection_arg = argparse.ArgumentParser(add_help=False)
    connection_arg.add_argument(
        '--address', '-a', type=str, default=None,
        help="IP address of console")
    connection_arg.add_argument(
        '--liveid', '-l',
        help='LiveID to poweron')

    server_args = argparse.ArgumentParser(add_help=False)
    server_args.add_argument(
        '--bind', '-b', default='127.0.0.1',
        help='Interface address to bind the server')
    server_args.add_argument(
        '--port', '-p', type=int, default=0,
        help='Port to bind to, defaults: (REST: 5557, REPL: 5558)')

    """Common argument for interactively choosing console to handle"""
    interactive_arg = argparse.ArgumentParser(add_help=False)
    interactive_arg.add_argument(
        '--interactive', '-i', action='store_true',
        help="Interactively choose console to connect to")

    """
    Define commands
    """
    subparsers = parser.add_subparsers(help='Available commands')
    # NOTE: Setting dest and required here for py3.6 compat
    subparsers.dest = 'command'
    subparsers.required = True

    """Discover"""
    subparsers.add_parser(Commands.Discover,
                          help='Discover console',
                          parents=[logging_args,
                                   connection_arg])

    """Power on"""
    subparsers.add_parser(
        Commands.PowerOn,
        help='Power on console',
        parents=[logging_args, connection_arg])

    """Power off"""
    poweroff_cmd = subparsers.add_parser(
        Commands.PowerOff,
        help='Power off console',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg])
    poweroff_cmd.add_argument(
        '--all', action='store_true',
        help="Power off all consoles")

    """Local REPL"""
    subparsers.add_parser(
        Commands.REPL,
        help='Local REPL (interactive console)',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg])

    """REPL server"""
    subparsers.add_parser(
        Commands.REPLServer,
        help='REPL server (interactive console)',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg,
                 server_args])

    """Fallout relay"""
    subparsers.add_parser(
        Commands.FalloutRelay,
        help='Fallout 4 Pip boy relay',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg])

    """Controller input"""
    subparsers.add_parser(
        Commands.GamepadInput,
        help='Send controller input to dashboard / apps',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg])

    """Text input"""
    subparsers.add_parser(
        Commands.TextInput,
        help='Client to use Text input functionality',
        parents=[logging_args, xbl_token_args,
                 interactive_arg, connection_arg])

    tui_cmd = subparsers.add_parser(
        Commands.TUI,
        help='TUI client - fancy :)',
        parents=[logging_args, xbl_token_args,
                 connection_arg])
    tui_cmd.add_argument(
        '--consoles', '-c', default=CONSOLES_FILE,
        help="Previously discovered consoles (json)")

    # FIXME: If possible include startup for REST server here too
    """
    REST server
    NOTE: Only argument parsing is handled in here,
          The actual start code is in a dedicated script.

          This is required due to gevent monkey patching for
          the FLASK web-framework to work.
    """
    subparsers.add_parser(
        Commands.RESTServer,
        help='REST server',
        parents=[logging_args, xbl_token_args, server_args])

    return parser.parse_args(args)


def handle_logging_setup(args):
    """
    Determine log level, logfile and special DEBUG_INCL_PACKETS
    via cmdline arguments.

    Args:
        args: ArgumentParser `Namespace`

    Returns:
         None
    """
    levels = [logging.WARNING, logging.INFO, logging.DEBUG, LOG_LEVEL_DEBUG_INCL_PACKETS]
    # Output level capped to number of levels
    log_level = levels[min(len(levels) - 1, args.verbose)]
    logging.basicConfig(level=log_level, format=LOG_FMT)
    logging.root.info('Set Loglevel: {0}'
                      .format(logging.getLevelName(log_level)))

    if log_level == LOG_LEVEL_DEBUG_INCL_PACKETS:
        logging.root.info('Removing previous logging StreamHandlers')
        while len(logging.root.handlers):
            del logging.root.handlers[0]

        logging.root.info('Using DEBUG_INCL_PACKETS logging')
        debugext_handler = logging.StreamHandler()
        debugext_handler.setFormatter(VerboseFormatter(LOG_FMT))
        logging.root.addHandler(debugext_handler)

    if args.logfile:
        logging.root.info('Set Logfile path: {0}'.format(args.logfile))
        file_handler = RotatingFileHandler(args.logfile, backupCount=2)
        file_handler.setLevel(log_level)
        file_handler.setFormatter(logging.Formatter(LOG_FMT))
        logging.root.addHandler(file_handler)


def do_authentication(token_filepath, do_refresh):
    """
    Shortcut for doing xbox live authentication (uses xbox-webapi-python lib).

    Args:
        token_filepath (str): Token filepath
        do_refresh (bool): Whether to refresh tokens

    Returns:
        AuthenticationManager: An authenticated instance

    Raises:
        AuthenticationException: If authentication failed
    """
    auth_mgr = AuthenticationManager.from_file(token_filepath)
    auth_mgr.authenticate(do_refresh=do_refresh)
    if do_refresh:
        auth_mgr.dump(token_filepath)

    return auth_mgr


def choose_console_interactively(console_list):
    """
    Choose a console to use via user-input

    Args:
        console_list (list): List of consoles to choose from

    Returns:
        None if choice was aborted, a desired console object otherwise
    """
    entry_count = len(console_list)
    LOGGER.debug('Offering console choices: {0}'.format(entry_count))

    print('Discovered consoles:')
    for idx, console in enumerate(console_list):
        print('  {0}: {1} {2} {3}'
              .format(idx, console.name, console.liveid, console.address))

    print('Enter \'x\' to abort')

    choices = [str(i) for i in range(entry_count)]
    choices.append('e')

    response = ''
    while response not in choices:
        response = input('Make your choice: ')
        if response == 'e':
            return None

    return console_list[int(response)]


def cli_discover_consoles(args):
    """
    Discover consoles
    """
    LOGGER.info('Sending discovery packets to IP: {0}'
                .format('IP: ' + args.address if args.address else '<MULTICAST>'))
    discovered = Console.discover(addr=args.address, timeout=1)

    if not len(discovered):
        LOGGER.error('No consoles discovered')
        sys.exit(ExitCodes.DiscoveryError)

    LOGGER.info('Discovered consoles ({0}): {1}'
                .format(len(discovered), ', '.join([str(c) for c in discovered])))

    if args.liveid:
        LOGGER.info('Filtering discovered consoles for LIVEID: {0}'
                    .format(args.liveid))
        discovered = [c for c in discovered if c.liveid == args.liveid]
    if args.address:
        LOGGER.info('Filtering discovered consoles for IP address: {0}'
                    .format(args.address))
        discovered = [c for c in discovered if c.address == args.address]

    return discovered


def main(command=None):
    """
    Main entrypoint
    """
    auth_manager = None
    repl_server_handle = None  # Used for Command.REPLServer

    if command:
        # Take passed command and append actual cmdline
        cmdline_arguments = sys.argv[1:]
        cmdline_arguments.insert(0, command)
    else:
        cmdline_arguments = None

    args = parse_arguments(cmdline_arguments)
    handle_logging_setup(args)

    LOGGER.debug('Parsed arguments: {0}'.format(args))

    command = args.command
    LOGGER.debug('Chosen command: {0}'.format(command))

    if command == Commands.RESTServer:
        LOGGER.info('Make sure you used the dedicated \'xbox-rest-server\' script'
                    ' to start the REST server!')

    elif 'interactive' in args and args.interactive and \
         (args.address or args.liveid):
        LOGGER.error('Flag \'--interactive\' is incompatible with'
                     ' providing an IP address (--address) or LiveID (--liveid) explicitly')
        sys.exit(ExitCodes.ArgParsingError)
    elif args.liveid and args.address:
        LOGGER.warning('You passed --address AND --liveid: Will only use that specific'
                       'combination!')
    elif command == Commands.PowerOff and args.all and (args.liveid or args.address):
        LOGGER.error('Poweroff with --all flag + explicitly provided LiveID / IP address makes no sense')
        sys.exit(ExitCodes.ArgParsingError)
    elif command == Commands.PowerOff and args.interactive and args.all:
        LOGGER.error('Combining args --all and --interactive not supported')
        sys.exit(ExitCodes.ArgParsingError)

    print('Xbox SmartGlass main client started')

    if command == Commands.RESTServer:
        """
        REST Server
        """

        if args.port == 0:
            LOGGER.info('No defaults provided, '
                        'Setting REST server port to {0}'.format(REST_DEFAULT_SERVER_PORT))
            args.port = REST_DEFAULT_SERVER_PORT

        print('Xbox Smartglass REST server started on {0}:{1}'.format(
            args.bind, args.port
        ))

        flask_app.token_file = args.tokens
        server = rest_pywsgi.WSGIServer((args.bind, args.port), flask_app)
        server.serve_forever()
        sys.exit(ExitCodes.OK)
    elif command == Commands.TUI:
        """
        Text user interface (powered by urwid)
        """
        # Removing stream handlers to not pollute TUI
        for h in [sh for sh in logging.root.handlers
                  if isinstance(sh, logging.StreamHandler)]:
            LOGGER.debug('Removing StreamHandler {0} from root logger'.format(h))
            logging.root.removeHandler(h)

        sys.exit(tui.run_tui(args.consoles, args.address,
                             args.liveid, args.tokens, args.refresh))

    elif 'tokens' in args:
        """
        Do Xbox live authentication
        """
        LOGGER.debug('Command {0} supports authenticated connection'.format(command))
        try:
            auth_manager = do_authentication(args.tokens, args.refresh)
        except AuthenticationException:
            LOGGER.exception('Authentication failed!')
            LOGGER.error("Please re-run xbox-authenticate to get a fresh set")
            sys.exit(ExitCodes.AuthenticationError)

    elif command == Commands.PowerOn:
        """
        Powering up console
        """
        if not args.liveid:
            LOGGER.error('No LiveID (--liveid) provided for power on!')
            sys.exit(ExitCodes.ArgParsingError)

        LOGGER.info('Sending poweron packet for LiveId: {0} to {1}'
                    .format(args.liveid,
                            'IP: ' + args.address if args.address else '<MULTICAST>'))
        Console.power_on(args.liveid, args.address, tries=10)
        sys.exit(0)

    """
    Discovery
    """
    discovered = cli_discover_consoles(args)

    if command == Commands.Discover:
        """
        Simply print discovered consoles
        """
        print("Discovered %d consoles: " % len(discovered))
        for console in discovered:
            print("  %s" % console)
        sys.exit(ExitCodes.OK)

    elif command == Commands.PowerOff and args.all:
        """
        Early call for poweroff --all
        """
        """Powering off all discovered consoles"""
        for c in discovered:
            print('Powering off console {0}'.format(c))
            c.power_off()
        sys.exit(ExitCodes.OK)

    """
    Choosing/filtering a console from the discovered ones
    """
    console = None
    if args.interactive:
        LOGGER.debug('Starting interactive console choice')
        console = choose_console_interactively(discovered)
    elif len(discovered) == 1:
        LOGGER.debug('Choosing sole console, no user interaction required')
        console = discovered[0]
    elif len(discovered) > 1:
        LOGGER.error(
            'More than one console was discovered and no exact'
            ' connection parameters were provided')

    if not console:
        LOGGER.error('Choosing a console failed!')
        sys.exit(ExitCodes.ConsoleChoice)

    LOGGER.info('Choosen target console: {0}'.format(console))

    LOGGER.debug('Setting console callbacks')
    console.on_device_status += \
        lambda x: LOGGER.info('Device status: {0}'.format(x))
    console.on_connection_state += \
        lambda x: LOGGER.info('Connection state: {0}'.format(x))
    console.on_pairing_state += \
        lambda x: LOGGER.info('Pairing state: {0}'.format(x))
    console.on_console_status += \
        lambda x: LOGGER.info('Console status: {0}'.format(x))
    console.on_timeout += \
        lambda x: LOGGER.error('Timeout occured!') or sys.exit(1)

    userhash = auth_manager.userinfo.userhash
    xtoken = auth_manager.xsts_token

    LOGGER.debug('Authentication info:')
    LOGGER.debug('Userhash: {0}'.format(userhash))
    LOGGER.debug('XToken: {0}'.format(xtoken))

    LOGGER.info('Attempting connection...')
    state = console.connect(userhash, xtoken.jwt)
    if state != ConnectionState.Connected:
        LOGGER.error('Connection failed! Console: {0}'.format(console))
        sys.exit(1)

    # FIXME: Waiting explicitly
    LOGGER.info('Connected to console: {0}'.format(console))
    LOGGER.debug('Waiting a second before proceeding...')
    console.wait(1)

    if command == Commands.PowerOff:
        """
        Power off (single console)
        """
        print('Powering off console {0}'.format(console))
        console.power_off()
        sys.exit(ExitCodes.OK)

    elif command == Commands.REPL or \
            command == Commands.REPLServer:

        banner = 'You are connected to the console @ {0}\n'\
                 .format(console.address)
        banner += 'Type in \'console\' to acccess the object\n'
        banner += 'Type in \'exit()\' to quit the application'

        scope_vars = {'console': console}

        if command == Commands.REPL:
            LOGGER.info('Starting up local REPL console')
            repl_local = InteractiveConsole(locals=scope_vars)
            repl_local.interact(banner)
        else:

            if args.port == 0:
                LOGGER.info('No defaults provided, '
                            'Setting REPL server port to {0}'.format(REPL_DEFAULT_SERVER_PORT))
                args.port = REPL_DEFAULT_SERVER_PORT

            startinfo = 'Starting up REPL server @ {0}:{1}'.format(args.bind, args.port)
            print(startinfo)
            LOGGER.info(startinfo)

            repl_server_handle = backdoor.BackdoorServer(
                listener=(args.bind, args.port),
                banner=banner,
                locals=scope_vars)

    elif command == Commands.FalloutRelay:
        """
        Fallout 4 relay
        """
        print('Starting Fallout 4 relay service...')
        console.add_manager(TitleManager)
        console.title.on_connection_info += fallout4_relay.on_connection_info
        console.start_title_channel(title_id=fallout4_relay.FALLOUT_TITLE_ID)
        print('Fallout 4 relay started')
    elif command == Commands.GamepadInput:
        """
        Gamepad input
        """
        print('Starting gamepad input handler...')
        console.add_manager(manager.InputManager)
        gamepad_input.input_loop(console)
    elif command == Commands.TextInput:
        """
        Text input
        """
        print('Starting text input handler...')
        console.add_manager(manager.TextManager)
        console.text.on_systemtext_configuration += text_input.on_text_config
        console.text.on_systemtext_input += functools.partial(text_input.on_text_input, console)
        console.text.on_systemtext_done += text_input.on_text_done

    LOGGER.debug('Installing gevent SIGINT handler')
    signal.signal(signal.SIGINT, lambda *a: console.protocol.stop())

    if repl_server_handle:
        LOGGER.debug('Starting REPL server protocol')

    LOGGER.debug('Starting console.protocol.serve_forever()')
    console.protocol.serve_forever()

    LOGGER.debug('Protocol serving exited')
    if repl_server_handle:
        LOGGER.debug('Stopping REPL server protocol')
        repl_server_handle.stop()


def main_discover():
    """Entrypoint for discover script"""
    main(Commands.Discover)


def main_poweron():
    """Entrypoint for poweron script"""
    main(Commands.PowerOn)


def main_poweroff():
    """Entrypoint for poweroff script"""
    main(Commands.PowerOff)


def main_repl():
    """Entrypoint for REPL script"""
    main(Commands.REPL)


def main_replserver():
    """Entrypoint for REPL server script"""
    main(Commands.REPLServer)


def main_falloutrelay():
    """Entrypoint for Fallout 4 relay script"""
    main(Commands.FalloutRelay)


def main_textinput():
    """Entrypoint for Text input script"""
    main(Commands.TextInput)


def main_gamepadinput():
    """Entrypoint for Gamepad input script"""
    main(Commands.GamepadInput)


def main_tui():
    """Entrypoint for TUI script"""
    main(Commands.TUI)


if __name__ == '__main__':
    main()



import pytest
from xbox.stump import json_model
from xbox.stump.enum import Message


def test_stump_response(stump_json):
    data = stump_json['response_recent_channels']
    msg = json_model.deserialize_stump_message(data)

    assert msg.msgid == 'xV5X1YCB.16'
    assert Message(msg.response) == Message.RECENT_CHANNELS
    assert isinstance(msg.params, list)


def test_tuner_lineups(stump_json):
    data = stump_json['response_tuner_lineups']
    msg = json_model.deserialize_stump_message(data)

    assert len(msg.params.providers) == 1
    provider = msg.params.providers[0]

    assert len(provider.foundChannels) == 19
    found_channel = provider.foundChannels[0]
    assert found_channel.channelNumber == 0
    assert found_channel.displayName == 'Das Erste HD'
    assert found_channel.channelId == '000021146A000301'

    assert len(provider.cqsChannels) == 7
    assert provider.cqsChannels[0] == '178442d3-2b13-e02b-9747-a3d4ebebcf62_PHOENIHD_23'

    assert str(provider.headendId) == '0a7fb88a-960b-c2e3-9975-7c86c5fa6c49'


def test_recent_channels(stump_json):
    data = stump_json['response_recent_channels']
    msg = json_model.deserialize_stump_message(data)

    assert len(msg.params) == 0
    # assert msg.params[0].channelNum == ''
    # assert msg.params[0].providerId == ''
    # assert msg.params[0].channelId == ''


def test_livetv_info(stump_json):
    data = stump_json['response_livetv_info']
    msg = json_model.deserialize_stump_message(data)

    assert msg.params.streamingPort == 10242
    assert msg.params.inHdmiMode is False
    assert msg.params.tunerChannelType == 'televisionChannel'
    assert msg.params.currentTunerChannelId == 'bb1ca492-232b-adfe-1f39-d010eabf179e_MSAHD_16'
    assert msg.params.currentHdmiChannelId == '731cd976-c1e9-6b95-4799-e6757d02cab1_3SATHD_1'
    assert msg.params.pauseBufferInfo is not None
    assert msg.params.pauseBufferInfo.Enabled is True
    assert msg.params.pauseBufferInfo.IsDvr is False
    assert msg.params.pauseBufferInfo.MaxBufferSize == 18000000000
    assert msg.params.pauseBufferInfo.BufferCurrent == 131688132168080320
    assert msg.params.pauseBufferInfo.BufferStart == 131688132168080320
    assert msg.params.pauseBufferInfo.BufferEnd == 131688151636700238
    assert msg.params.pauseBufferInfo.CurrentTime == 131688151636836518
    assert msg.params.pauseBufferInfo.Epoch == 0


def test_headend_info(stump_json):
    data = stump_json['response_headend_info']
    msg = json_model.deserialize_stump_message(data)

    assert msg.params.providerName == 'Sky Deutschland'
    assert str(msg.params.headendId) == '516b9ea7-5292-97ec-e7d4-f843fab6d392'
    assert msg.params.blockExplicitContentPerShow is False
    assert msg.params.dvrEnabled is False
    assert msg.params.headendLocale == 'de-DE'
    assert msg.params.streamingPort == 10242
    assert msg.params.preferredProvider == '29045393'

    assert len(msg.params.providers) == 2
    provider = msg.params.providers[0]
    assert provider.providerName == 'Sky Deutschland'
    assert provider.filterPreference == 'ALL'
    assert str(provider.headendId) == '516b9ea7-5292-97ec-e7d4-f843fab6d392'
    assert provider.source == 'hdmi'
    assert provider.titleId == '162615AD'
    assert provider.canStream is False


def test_configuration(stump_json):
    data = stump_json['response_configuration']
    msg = json_model.deserialize_stump_message(data)

    assert len(msg.params) == 4
    device_config = msg.params[0]
    assert device_config.device_brand == 'Samsung'
    assert device_config.device_id == '0'
    assert device_config.device_type == 'tv'
    assert isinstance(device_config.buttons, dict) is True


@pytest.mark.skip
def test_ensure_streaming_started(stump_json):
    data = stump_json['response_ensure_streaming_started']
    msg = json_model.deserialize_stump_message(data)

    assert msg.params.currentChannelId == ''
    assert msg.params.source == ''
    assert msg.params.streamingPort == 0
    assert msg.params.tunerChannelType == ''
    assert msg.params.userCanViewChannel == ''


def test_app_channel_lineups(stump_json):
    data = stump_json['response_appchannel_lineups']
    msg = json_model.deserialize_stump_message(data)

    assert len(msg.params) == 4
    provider = msg.params[0]
    assert provider.id == 'LiveTvHdmiProvider'
    assert provider.providerName == 'OneGuide'
    assert provider.titleId == '00000000'
    assert provider.primaryColor == 'ff107c10'
    assert provider.secondaryColor == 'ffebebeb'
    assert len(provider.channels) == 0
    # channel = provider.channels[0]
    # assert channel.name == ''
    # assert channel.id == ''


def test_send_key(stump_json):
    data = stump_json['response_sendkey']
    msg = json_model.deserialize_stump_message(data)

    assert msg.params is True



import pytest
from xbox.sg.protocol import _fragment_connect_request, FragmentError
from xbox.sg.protocol import ChannelManager, ChannelError
from xbox.sg.protocol import SequenceManager
from xbox.sg.protocol import FragmentManager


def test_fragment_connect_request(crypto):
    from uuid import UUID
    from xbox.sg.enum import PublicKeyType

    uuid = UUID('de305d54-75b4-431b-adb2-eb6b9e546014')
    pubkey = 64 * b'\xFF'
    userhash = '0123456789'
    auth_token = (898 * 'A') + (500 * 'B')

    fragments = _fragment_connect_request(crypto, uuid, PublicKeyType.EC_DH_P256,
                                          pubkey, userhash, auth_token)
    # request_num != 0
    fragments_dup = _fragment_connect_request(crypto, uuid, PublicKeyType.EC_DH_P256,
                                              pubkey, userhash, auth_token,
                                              request_num=3)

    # Total length (userhash + token) is too small for fragmentation
    with pytest.raises(FragmentError):
        _fragment_connect_request(crypto, uuid, PublicKeyType.EC_DH_P256,
                                  pubkey, userhash, 898 * 'A')

    assert len(fragments) == 2
    assert fragments[0].unprotected_payload.sg_uuid == uuid
    assert fragments[0].unprotected_payload.public_key == pubkey
    assert isinstance(fragments[0].unprotected_payload.iv, bytes) is True
    assert fragments[0].protected_payload.userhash == userhash
    assert fragments[0].protected_payload.jwt == 898 * 'A'
    assert fragments[0].protected_payload.connect_request_num == 0
    assert fragments[0].protected_payload.connect_request_group_start == 0
    assert fragments[0].protected_payload.connect_request_group_end == 2

    assert fragments[1].unprotected_payload.sg_uuid == uuid
    assert fragments[1].unprotected_payload.public_key == pubkey
    assert isinstance(fragments[1].unprotected_payload.iv, bytes) is True
    assert fragments[1].protected_payload.userhash == ''
    assert fragments[1].protected_payload.jwt == 500 * 'B'
    assert fragments[1].protected_payload.connect_request_num == 1
    assert fragments[1].protected_payload.connect_request_group_start == 0
    assert fragments[1].protected_payload.connect_request_group_end == 2

    assert fragments_dup[0].protected_payload.connect_request_num == 3
    assert fragments_dup[0].protected_payload.connect_request_group_start == 3
    assert fragments_dup[0].protected_payload.connect_request_group_end == 5

    assert fragments_dup[1].protected_payload.connect_request_num == 4
    assert fragments_dup[1].protected_payload.connect_request_group_start == 3
    assert fragments_dup[1].protected_payload.connect_request_group_end == 5


def test_sequence_manager():
    mgr = SequenceManager()

    seq_num = 0
    for i in range(5):
        seq_num = mgr.next_sequence_num()

    for i in range(1, 23):
        mgr.add_received(i)

    for i in range(1, 12):
        mgr.add_processed(i)

    for i in range(1, 7):
        mgr.add_rejected(i)

    # Trying to set already existing values
    mgr.add_received(4)
    mgr.add_processed(5)
    mgr.add_rejected(6)

    mgr.low_watermark = 89
    # Setting smaller sequence than already set
    mgr.low_watermark = 12

    assert mgr.received == list(range(1, 23))
    assert mgr.processed == list(range(1, 12))
    assert mgr.rejected == list(range(1, 7))

    assert mgr.low_watermark == 89
    assert seq_num == 5


def test_channel_manager(packets, crypto):
    from xbox.sg.enum import ServiceChannel
    from xbox.sg.packer import unpack

    start_channel_resp = unpack(packets['start_channel_response'], crypto)
    mgr = ChannelManager()

    request_id = 0
    while request_id != start_channel_resp.protected_payload.channel_request_id:
        request_id = mgr.get_next_request_id(ServiceChannel.SystemInputTVRemote)

    service_channel = mgr.handle_channel_start_response(start_channel_resp)

    assert service_channel == ServiceChannel.SystemInputTVRemote

    # Deleting channel references
    mgr.reset()

    with pytest.raises(ChannelError):
        mgr.get_channel(123)

    with pytest.raises(ChannelError):
        mgr.get_channel_id(ServiceChannel.SystemInputTVRemote)

    with pytest.raises(ChannelError):
        mgr.handle_channel_start_response(start_channel_resp)

    assert mgr.get_channel(0) == ServiceChannel.Core
    assert mgr.get_channel(0x1000000000000000) == ServiceChannel.Ack

    assert mgr.get_channel_id(ServiceChannel.Core) == 0
    assert mgr.get_channel_id(ServiceChannel.Ack) == 0x1000000000000000


def test_fragment_manager_json(json_fragments):
    expected_result = {'response': 'GetConfiguration',
                       'msgid': 'xV5X1YCB.13',
                       'params': [
                           {'device_id': '0',
                            'device_type': 'tv',
                            'buttons': {
                                'btn.back': 'Back',
                                'btn.up': 'Up',
                                'btn.red': 'Red',
                                'btn.page_down': 'Page Down',
                                'btn.ch_down': 'Channel Down',
                                'btn.func_c': 'Label C',
                                'btn.format': 'Format',
                                'btn.digit_2': '2',
                                'btn.func_a': 'Label A',
                                'btn.digit_7': '7',
                                'btn.last': 'Last',
                                'btn.input': 'Input',
                                'btn.fast_fwd': 'FFWD',
                                'btn.menu': 'Menu',
                                'btn.replay': 'Skip REV',
                                'btn.power': 'Power',
                                'btn.left': 'Left',
                                'btn.blue': 'Blue',
                                'btn.vol_down': 'Volume Down',
                                'btn.green': 'Green',
                                'btn.digit_4': '4',
                                'btn.digit_9': '9',
                                'btn.play': 'Play',
                                'btn.page_up': 'Page Up',
                                'btn.func_b': 'Label B',
                                'btn.power_off': 'Off',
                                'btn.vol_mute': 'Mute',
                                'btn.record': 'Record',
                                'btn.subtitle': 'Subtitle',
                                'btn.rewind': 'Rewind',
                                'btn.exit': 'Exit',
                                'btn.down': 'Down',
                                'btn.sap': 'Sap',
                                'btn.yellow': 'Yellow',
                                'btn.func_d': 'Label D',
                                'btn.info': 'Info',
                                'btn.digit_5': '5',
                                'btn.digit_3': '3',
                                'btn.digit_0': '0',
                                'btn.skip_fwd': 'Skip FWD',
                                'btn.delimiter': 'Delimiter',
                                'btn.right': 'Right',
                                'btn.vol_up': 'Volume Up',
                                'btn.ch_up': 'Channel Up',
                                'btn.digit_8': '8',
                                'btn.digit_6': '6',
                                'btn.guide': 'Guide',
                                'btn.stop': 'Stop',
                                'btn.select': 'Select',
                                'btn.power_on': 'On',
                                'btn.ch_enter': 'Enter',
                                'btn.digit_1': '1',
                                'btn.pause': 'Pause',
                                'btn.dvr': 'Recordings'
                            }},
                           {'device_id': '1',
                            'device_brand': 'Sky Deutschland',
                            'device_type': 'stb',
                            'buttons': {
                                'btn.back': 'Back',
                                'btn.red': 'Red',
                                'btn.up': 'Up',
                                'btn.ch_down': 'Channel Down',
                                'btn.format': 'Format',
                                'btn.digit_2': '2',
                                'btn.digit_7': '7',
                                'btn.last': 'Last',
                                'btn.fast_fwd': 'FFWD',
                                'btn.menu': 'Menu',
                                'btn.power': 'Power',
                                'btn.left': 'Left',
                                'btn.blue': 'Blue',
                                'btn.vol_down': 'Volume Down',
                                'btn.green': 'Green',
                                'btn.digit_4': '4',
                                'btn.digit_9': '9',
                                'btn.play': 'Play',
                                'btn.vol_mute': 'Mute',
                                'btn.record': 'Record',
                                'btn.rewind': 'Rewind',
                                'btn.exit': 'Exit',
                                'btn.down': 'Down',
                                'btn.yellow': 'Yellow',
                                'btn.info': 'Info',
                                'btn.digit_5': '5',
                                'btn.digit_3': '3',
                                'btn.digit_0': '0',
                                'btn.live': 'Live',
                                'btn.vol_up': 'Volume Up',
                                'btn.right': 'Right',
                                'btn.ch_up': 'Channel Up',
                                'btn.digit_8': '8',
                                'btn.digit_6': '6',
                                'btn.guide': 'Guide',
                                'btn.stop': 'Stop',
                                'btn.select': 'Select',
                                'btn.digit_1': '1',
                                'btn.pause': 'Pause',
                                'btn.dvr': 'Recordings'}
                            },
                           {'device_id': 'tuner',
                            'device_type': 'tuner',
                            'buttons': {
                                'btn.play': 'PLAY',
                                'btn.pause': 'PAUSE',
                                'btn.seek': 'SEEK'
                            }
                            }
                       ]}

    mgr = FragmentManager()

    for msg in json_fragments:
        msg1 = mgr.reassemble_json(msg)

    for msg in reversed(json_fragments):
        msg2 = mgr.reassemble_json(msg)

    # Deliver a fragment twice
    mgr.reassemble_json(json_fragments[0])
    mgr.reassemble_json(json_fragments[1])
    mgr.reassemble_json(json_fragments[2])
    mgr.reassemble_json(json_fragments[2])
    msg3 = mgr.reassemble_json(json_fragments[3])

    # Incomplete fragments
    for msg in reversed(json_fragments[:-1]):
        msg4 = mgr.reassemble_json(msg)

    with pytest.raises(KeyError):
        mgr.reassemble_json({'not': 'all', 'required': 'fields'})

    assert msg1 == expected_result
    assert msg2 == expected_result
    assert msg3 == expected_result
    assert msg4 is None


def test_fragment_manager_fragment_messages(packets, crypto):
    from xbox.sg.packer import unpack

    fragments = [
        unpack(packets['fragment_media_state_0'], crypto),
        unpack(packets['fragment_media_state_1'], crypto),
        unpack(packets['fragment_media_state_2'], crypto)
    ]

    mgr = FragmentManager()
    assert mgr.reassemble_message(fragments.pop()) is None
    assert mgr.reassemble_message(fragments.pop()) is None
    msg = mgr.reassemble_message(fragments.pop())
    assert msg is not None

    assert msg.aum_id == 'Microsoft.BlurayPlayer_8wekyb3d8bbwe!Xbox.BlurayPlayer.Application'
    assert msg.max_seek == 50460000
    assert len(msg.asset_id) == 2184



"""
Terminal UI Smartglass Client

Supported functions: Poweron/off, launch title, gamepad input and entering text.
Additional shows console status (active titles, OS version, locale) and media state.
"""
import json
import urwid
import logging
from binascii import hexlify
from collections import deque

import gevent
import gevent.select
import gevent.signal

from xbox.webapi.scripts.tui import WebAPIDisplay

from xbox.scripts import ExitCodes
from xbox.sg.console import Console
from xbox.sg.enum import DeviceStatus, GamePadButton, MediaPlaybackStatus
from xbox.sg.manager import InputManager, TextManager, MediaManager

from construct.lib import containers
containers.setGlobalPrintFullStrings(True)


class ControllerRemote(urwid.Filler):
    keymap = {
        'tab': GamePadButton.View,
        '<': GamePadButton.Menu,
        '#': GamePadButton.Nexu,
        'up': GamePadButton.DPadUp,
        'down': GamePadButton.DPadDown,
        'left': GamePadButton.DPadLeft,
        'right': GamePadButton.DPadRight,
        'a': GamePadButton.PadA,
        'b': GamePadButton.PadB,
        'x': GamePadButton.PadX,
        'y': GamePadButton.PadY
    }
    text = 'Use keyboard to send controller input'

    def __init__(self, app, console, original_widget):
        self.app = app
        self.console = console

        super(ControllerRemote, self).__init__(original_widget, 'top')

    def keypress(self, size, key):
        if key in self.keymap:
            button = self.keymap[key]
            self.console.gamepad_input(button)
            self.console.gamepad_input(GamePadButton.Clear)
        elif key in ('q', 'Q'):
            return key
        else:
            return super(ControllerRemote, self).keypress(size, key)


class QuestionBox(urwid.Filler):
    def __init__(self, edit_widget, callback, **kwargs):
        super(QuestionBox, self).__init__(edit_widget, **kwargs)
        self.callback = callback

    def keypress(self, size, key):
        if key != 'enter':
            return super(QuestionBox, self).keypress(size, key)
        else:
            self.callback(self.original_widget.edit_text)


class TextInput(urwid.Filler):
    def __init__(self, app, edit_widget, console, **kwargs):
        super(TextInput, self).__init__(edit_widget, **kwargs)
        self.app = app
        self.console = console

    def keypress(self, size, key):
        if key != 'enter':
            ret = super(TextInput, self).keypress(size, key)
            self.console.send_systemtext_input(self.original_widget.edit_text)
            return ret
        else:
            self.console.finish_text_input()
            self.app.return_to_details_menu()


class MediaProgressbar(urwid.ProgressBar):
    def __init__(self, normal, complete):
        super(MediaProgressbar, self).__init__(normal, complete)
        self._current_text = 'No media playing'

    def get_text(self):
        return self._current_text

    def update_from_state(self, state):
        if not state or state.playback_status in (MediaPlaybackStatus.Stopped, MediaPlaybackStatus.Closed):
            self._current_text = 'No media playing'
            self.set_completion(0.0)
        else:
            pos_seconds = state.position / 10000000
            total_seconds = state.media_end / 10000000
            self._current_text = '{pmin:02d}:{psec:02d} / {tmin:02d}:{tsec:02d}'.format(
                pmin=int(pos_seconds // 60),
                psec=int(pos_seconds % 60),
                tmin=int(total_seconds // 60),
                tsec=int(total_seconds % 60)
            )
            if state.media_end > 0:
                self.done = state.media_end
            self.set_completion(state.position)


class ConsoleView(urwid.Frame):
    def __init__(self, app, console, **kwargs):
        self.app = app
        self.console = console

        self.device_info = urwid.LineBox(urwid.Text('Not available'), 'Device info')
        self.status = urwid.LineBox(urwid.Text('Not available'), 'Console status')
        self.media_text = urwid.Text('Not available')
        self.media_progress = MediaProgressbar('pg normal', 'pg complete')
        media_state_pile = urwid.Pile([self.media_text, self.media_progress])
        self.media_state = urwid.LineBox(media_state_pile, 'Media State')

        self.pile = urwid.Pile([self.device_info, self.status, self.media_state])
        self.filler = urwid.Filler(self.pile, valign='top')

        self.update_device_info()
        self.console.on_timeout += self.update_device_info
        self.console.on_pairing_state += lambda _: self.update_device_info()
        self.console.on_connection_state += lambda _: self.update_device_info()
        self.console.on_active_surface += lambda _: self.update_device_info()
        self.console.on_console_status += self.on_console_status
        self.console.media.on_media_state += self.on_media_state
        self.console.text.on_systemtext_configuration += lambda _: self.app.view_text_input_overlay(self.console)
        self.console.text.on_systemtext_input += lambda _: None  # Update text overlay with input text
        self.console.text.on_systemtext_done += lambda _: self.app.return_to_details_menu()

        super(ConsoleView, self).__init__(self.filler, **kwargs)

    def update_device_info(self):
        text = 'Name: {c.name:<15}\nAddress: {c.address:<15}\nLiveID: {c.liveid:<15}\n' \
               'UUID: {c.uuid}\n\n'.format(c=self.console)

        text += 'Connection State: {:<15}\n'.format(self.console.connection_state.name)
        text += 'Pairing State: {:<15}\n'.format(self.console.pairing_state.name)
        # text += 'Active Surface: {}\n'.format(ActiveSurfaceType[self.console.active_surface.surface_type])
        text += 'Shared secret: {hex_secret}'.format(
            hex_secret=hexlify(self.console._crypto.shared_secret).decode('utf-8')
        )
        self.device_info.original_widget.set_text(text)

    def on_console_status(self, console_status):
        if not console_status:
            self.status.original_widget.set_text('Not available')
            return

        text = \
            'LiveTV Provider: {status.live_tv_provider:<10}' \
            'Locale: {status.locale:<10}' \
            'OS: {status.major_version}.{status.minor_version}.{status.build_number}\n'.format(
                status=console_status
            )
        for title in console_status.active_titles:
            text += '{focus} {title.aum} (0x{title.title_id:08x}) [{location}]\n'.format(
                focus='*' if title.disposition.has_focus else ' ',
                title=title, location=title.disposition.title_location.name
            )
        self.status.original_widget.set_text(text)

    def on_media_state(self, state):
        if not state:
            self.media_text.set_text('Not available')
            self.media_progress.update_from_state(state)
            return

        text = \
            'Title: {state.aum_id:<20} (0x{state.title_id:08x}) AssetId: {state.asset_id:<25}\n' \
            'MediaType: {media_type:<25} SoundLevel: {sound_level:<25}\n' \
            'Playback: {playback_status:<25}\n'.format(
                state=state, media_type=state.media_type.name, sound_level=state.sound_level.name,
                playback_status=state.playback_status.name
            )
        for metadata in state.metadata:
            text += '{metadata.name}: {metadata.value}\n'.format(metadata=metadata)
        if state.playback_status in (MediaPlaybackStatus.Stopped, MediaPlaybackStatus.Closed):
            self.media_text.set_text('Not available')
        else:
            self.media_text.set_text(text)
        self.media_progress.update_from_state(state)

    def keypress(self, size, key):
        if key in ('c', 'C'):
            self.app.view_commands_menu(self.console)
        else:
            return super(ConsoleView, self).keypress(size, key)


class ConsoleButton(urwid.Button):
    focus_map = {
        None: 'selected',
        'connected': 'connected selected'
    }

    def __init__(self, app, console):
        super(ConsoleButton, self).__init__('')
        self.app = app

        self.console = console
        self.console.add_manager(InputManager)
        self.console.add_manager(MediaManager)
        self.console.add_manager(TextManager)
        self.console.on_connection_state += lambda _: self.refresh()
        self.console.on_console_status += lambda _: self.refresh()
        self.console.on_device_status += lambda _: self.refresh()

        urwid.connect_signal(self, 'click', self.callback)
        self.textwidget = urwid.AttrWrap(urwid.SelectableIcon('', cursor_position=0), None)
        self._w = urwid.AttrMap(self.textwidget, None, self.focus_map)
        self.refresh()

    def callback(self, *args):
        if not self.connect():
            return

        self.app.view_details_menu(self.console)

    def connect(self):
        if self.console.connected:
            return True

        if not self.console.available:
            self.app.view_msgbox('Console unavailable, try refreshing')
            return False

        state = self.console.connect(
            userhash=self.app.auth_mgr.userinfo.userhash,
            xsts_token=self.app.auth_mgr.xsts_token.jwt
        )

        if not self.console.connected:
            self.app.view_msgbox('Connection failed! State: {}'.format(state))
            return False

        return True

    def disconnect(self):
        if self.console.connected:
            self.console.disconnect()

    def refresh(self):
        text = ' {c.name:<20}{c.address:<20}{c.liveid:<20}{ds:<20}'.format(
            c=self.console, ds='{}, {}'.format(self.console.device_status.name, self.console.connection_state.name)
        )
        self.textwidget.set_text(text)

        if self.console.connected:
            self.textwidget.set_attr('connected')
        else:
            self.textwidget.set_attr(None)

    def keypress(self, size, key):
        if key in ('p', 'P'):
            self.console.power_on()
        elif key in ('c', 'C'):
            self.connect()
        elif key in ('d', 'D'):
            self.disconnect()
        else:
            return super(ConsoleButton, self).keypress(size, key)


class ConsoleList(urwid.Frame):
    def __init__(self, app, consoles, header, footer):
        walker = urwid.SimpleFocusListWalker([])
        listbox = urwid.ListBox(walker)

        frame = urwid.Frame(listbox, header=urwid.Text(' {0:<20}{1:<20}{2:<20}{3:<20}'.format(
            'Name', 'IP Address', 'Live ID', 'Status'
        )))
        view = urwid.LineBox(frame, 'Consoles')

        self.walker = walker
        self.app = app
        self.consoles = consoles
        super(ConsoleList, self).__init__(view, header=header, footer=footer)
        self.walker[:] = [ConsoleButton(self.app, c) for c in self.consoles]

    def refresh(self):
        gevent.spawn(self._refresh)

    def _refresh(self):
        discovered = Console.discover(blocking=True)

        liveids = [d.liveid for d in discovered]
        for i, c in enumerate(self.consoles):
            if c.liveid in liveids:
                # Refresh existing entries
                idx = liveids.index(c.liveid)
                if c.device_status != discovered[idx].device_status:
                    self.consoles[i] = discovered[idx]
                del discovered[idx]
                del liveids[idx]
            elif c.liveid not in liveids:
                # Set unresponsive consoles to Unavailable
                self.consoles[i].device_status = DeviceStatus.Unavailable

        # Add newly discovered consoles
        self.consoles.extend(discovered)

        # Update the consolelist view
        self.walker[:] = [ConsoleButton(self.app, c) for c in self.consoles]

    def keypress(self, size, key):
        if key in ('r', 'R'):
            self.refresh()
        else:
            return super(ConsoleList, self).keypress(size, key)


class CommandList(urwid.SimpleFocusListWalker):
    def __init__(self, app, console):
        commands = [
            ('Launch title', self._launch_title),
            ('Controller remote', self._controller_remote),
            ('Disconnect', self._disconnect),
            ('Power off', self._power_off)
        ]
        self.app = app
        self.console = console
        super(CommandList, self).__init__([CommandButton(text, func) for text, func in commands])

    def _launch_title(self):
        self.app.view_launch_title_textbox(self.__launch_title)

    def __launch_title(self, uri):
        self.console.launch_title(uri)
        self.app.return_to_details_menu()

    def _controller_remote(self):
        self.app.view_controller_remote_overlay(self.console)

    def _disconnect(self):
        self.console.disconnect()
        self.app.return_to_main_menu()

    def _power_off(self):
        self.console.power_off()
        self.app.return_to_main_menu()


class CommandButton(urwid.Button):
    focus_map = {
        None: 'selected',
    }

    def __init__(self, text, func):
        super(CommandButton, self).__init__('')
        urwid.connect_signal(self, 'click', self.callback)
        self.text = text
        self.func = func
        self.textwidget = urwid.AttrWrap(urwid.SelectableIcon(' {}'.format(self.text), cursor_position=0), None)
        self._w = urwid.AttrMap(self.textwidget, None, self.focus_map)

    def callback(self, *args):
        self.func()


class UrwidLogHandler(logging.Handler):
    def __init__(self, callback):
        super(UrwidLogHandler, self).__init__()
        self.callback = callback

    def emit(self, record):
        try:
            self.callback(record)
        except Exception:
            self.handleError(record)


class LogListBox(urwid.ListBox):
    def __init__(self, app, size=10000):
        self.app = app
        self.size = size
        self.entries = urwid.SimpleFocusListWalker([])

        self.handler = UrwidLogHandler(self._log_callback)
        self.handler.setFormatter(app.log_fmt)
        logging.root.addHandler(self.handler)
        logging.root.setLevel(app.log_level)
        super(LogListBox, self).__init__(self.entries)

    def _log_callback(self, record):
        self.entries.append(LogButton(self.app, self.handler.format(record), record))
        if self.focus_position == len(self.entries) - 2:
            self.focus_position += 1

        if len(self.entries) > self.size:
            self.entries[:] = self.entries[len(self.entries) - self.size:]

    def keypress(self, size, key):
        # Prevents opening the log window multiple times
        if key in ('l', 'L'):
            pass
        else:
            return super(LogListBox, self).keypress(size, key)


class LogButton(urwid.Button):
    focus_map = {
        None: 'selected',
    }

    def __init__(self, app, text, record):
        super(LogButton, self).__init__('')
        self.app = app
        self.text = text
        self.record = record

        urwid.connect_signal(self, 'click', self._click)
        self.textwidget = urwid.AttrWrap(urwid.SelectableIcon(' {}'.format(self.text), cursor_position=0), None)
        self._w = urwid.AttrMap(self.textwidget, None, self.focus_map)

    def _click(self, *args):
        if hasattr(self.record, '_msg'):
            self.app.view_scrollable_overlay(repr(self.record._msg), "Message details", width=80)


class SGDisplay(object):
    palette = [
        ('header', 'yellow', 'dark blue', 'standout'),
        ('listbar', 'light cyan,bold', 'default'),
        ('selected', 'black', 'light gray'),

        # footer
        ('foot', 'dark cyan', 'dark blue', 'bold'),
        ('key', 'light cyan', 'dark blue', 'underline'),

        # status
        ('connected', 'dark green', ''),
        ('connected selected', 'black', 'dark green'),

        # progressbar
        ('pg normal', 'white', 'black', 'standout'),
        ('pg complete', 'white', 'dark magenta'),
        ('pg smooth', 'dark magenta', 'black'),

        ('caption', 'yellow,bold', 'dark cyan'),
        ('prompt', 'white', 'dark cyan'),
        ('dialog', 'white', 'dark cyan'),
        ('button', 'white', 'dark cyan'),
        ('button selected', 'white', 'dark cyan'),
    ]

    header_text = ('header', [
        "Xbox Smartglass"
    ])

    footer_main_text = ('foot', [
        ('key', 'R:'), "reload  ",
        ('key', 'C:'), "connect ",
        ('key', 'D:'), "disconnect ",
        ('key', 'P:'), "poweron ",
        ('key', 'L:'), "view log ",
        ('key', 'Q:'), "quit  "
    ])

    footer_console_text = ('foot', [
        ('key', 'C:'), "commands ",
        ('key', 'L:'), "view log ",
        ('key', 'Q:'), "quit "
    ])

    footer_log_text = ('foot', [
        ('key', 'ENTER:'), "show details ",
        ('key', 'Q:'), "quit "
    ])

    log_fmt = logging.Formatter(logging.BASIC_FORMAT)
    log_level = logging.DEBUG

    def __init__(self, consoles, auth_mgr):
        self.header = urwid.AttrMap(urwid.Text(self.header_text), 'header')
        footer = urwid.AttrMap(urwid.Text(self.footer_main_text), 'foot')

        self.loop = None
        self.consoles = ConsoleList(self, consoles, self.header, footer)
        self.auth_mgr = auth_mgr
        self.log = LogListBox(self)

        self.view_stack = []

    def push_view(self, sender, view):
        self.view_stack.append(view)
        self.loop.widget = view
        self.loop.draw_screen()

    def pop_view(self, sender):
        if len(self.view_stack) > 1:
            top_widget = self.view_stack.pop()
            if hasattr(top_widget, 'close_view'):
                top_widget.close_view(sender)

            self.loop.widget = self.view_stack[-1]
            self.loop.draw_screen()
        else:
            self.do_quit()

    def view_main_menu(self):
        self.push_view(self, self.consoles)

    def view_details_menu(self, console):
        footer = urwid.AttrMap(urwid.Text(self.footer_console_text), 'foot')
        frame = ConsoleView(self, console, header=self.header, footer=footer)
        self.push_view(self, frame)

    def view_commands_menu(self, console):
        bottom = self.view_stack[-1]
        commands = urwid.ListBox(CommandList(self, console))
        top = urwid.LineBox(commands, 'Commands')
        overlay = urwid.Overlay(top, bottom, 'center', ('relative', 25), 'middle', ('relative', 75))
        self.push_view(self, overlay)

    def view_controller_remote_overlay(self, console):
        self.return_to_details_menu()
        bottom = self.view_stack[-1]
        edit = urwid.Edit('Press Q to quit\n'
                          'View: <tab>, Menu: <, Nexus: #\n'
                          'DPad: <Arrow keys>, Button: A-B-X-Y\n')
        view = ControllerRemote(self, console, edit)
        top = urwid.LineBox(view, title='Controller Remote')
        overlay = urwid.Overlay(top, bottom,
                                'center', ('relative', 25), 'middle', ('relative', 25))
        self.push_view(self, overlay)

    def view_text_input_overlay(self, console):
        self.return_to_details_menu()
        bottom = self.view_stack[-1]
        edit = urwid.Edit('Enter text\n'
                          'Press ENTER to send\n')
        view = TextInput(self, edit, console)
        top = urwid.LineBox(view, title='SystemText Session')
        overlay = urwid.Overlay(top, bottom,
                                'center', ('relative', 25), 'middle', ('relative', 25))
        self.push_view(self, overlay)

    def view_launch_title_textbox(self, callback):
        self.return_to_details_menu()
        bottom = self.view_stack[-1]
        edit = urwid.Edit('Press ENTER to send uri\n')
        question_box = QuestionBox(edit, callback)

        top = urwid.LineBox(question_box, title='Enter launch uri')
        overlay = urwid.Overlay(top, bottom,
                                'center', ('relative', 25), 'middle', ('relative', 25))
        self.push_view(self, overlay)

    def view_msgbox(self, msg, title='Error', width=25, height=75):
        bottom = self.view_stack[-1]
        text = urwid.Text(msg)
        button = urwid.Button('OK')
        button._label.align = 'center'
        pad_button = urwid.Padding(button, 'center', ('relative', width * 2))
        pile = urwid.Pile([text, pad_button])
        box = urwid.LineBox(pile, title)
        top = urwid.Filler(box, 'top')
        overlay = urwid.Overlay(top, bottom, 'center', ('relative', width), 'middle', ('relative', height))
        urwid.connect_signal(button, 'click', lambda _: self.pop_view(self))
        self.push_view(self, overlay)

    def view_scrollable_overlay(self, msg, title='Error', width=25, height=75):
        bottom = self.view_stack[-1]
        walker = urwid.SimpleFocusListWalker([urwid.Text(l) for l in msg.split('\n')])
        text = urwid.ListBox(walker)
        line = urwid.LineBox(text, title)
        overlay = urwid.Overlay(line, bottom, 'center', ('relative', width), 'middle', ('relative', height))
        self.push_view(self, overlay)

    def view_log(self):
        header = urwid.AttrMap(urwid.Text(self.header_text), 'header')
        footer = urwid.AttrMap(urwid.Text(self.footer_log_text), 'foot')
        frame = urwid.Frame(self.log, header=header, footer=footer)
        self.push_view(self, frame)

    def return_to_main_menu(self):
        while len(self.view_stack) > 1:
            self.pop_view(self)

    def return_to_details_menu(self):
        while len(self.view_stack) > 2:
            self.pop_view(self)

    def do_quit(self):
        raise urwid.ExitMainLoop()

    def run(self):
        self.loop = urwid.MainLoop(
            urwid.SolidFill('x'),
            handle_mouse=False,
            palette=self.palette,
            unhandled_input=self.unhandled_input,
            event_loop=GeventLoop()
        )

        self.loop.set_alarm_in(0.0001, lambda *args: self.view_main_menu())
        self.consoles.refresh()
        self.loop.run()

    def unhandled_input(self, input):
        if input in ('q', 'Q', 'esc'):
            self.pop_view(self)
        elif input in ('l', 'L'):
            self.view_log()


# https://github.com/what-studio/urwid-geventloop
class GeventLoop(object):
    def __init__(self):
        super(GeventLoop, self).__init__()
        self._completed_greenlets = deque()
        self._idle_callbacks = []
        self._idle_event = gevent.event.Event()

    def _greenlet_completed(self, greenlet):
        self._completed_greenlets.append(greenlet)
        self._idle_event.set()

    # alarm
    def alarm(self, seconds, callback):
        greenlet = gevent.spawn_later(seconds, callback)
        greenlet.link(self._greenlet_completed)
        return greenlet

    def remove_alarm(self, handle):
        if not handle.dead or handle.ready():
            handle.kill()
            return True
        return False

    # file
    def _watch_file(self, fd, callback):
        while True:
            gevent.select.select([fd], [], [])
            gevent.spawn(callback).link(self._greenlet_completed)

    def watch_file(self, fd, callback):
        greenlet = gevent.spawn(self._watch_file, fd, callback)
        greenlet.link(self._greenlet_completed)
        return greenlet

    def remove_watch_file(self, handle):
        handle.kill()
        return True

    # idle
    def enter_idle(self, callback):
        self._idle_callbacks.append(callback)
        return callback

    def remove_enter_idle(self, handle):
        try:
            self._idle_callbacks.remove(handle)
        except KeyError:
            return False
        return True

    # signal
    def set_signal_handler(self, signum, handler):
        gevent.signal.signal(signum, handler)

    def run(self):
        try:
            while True:
                for callback in self._idle_callbacks:
                    callback()
                while len(self._completed_greenlets) > 0:
                    self._completed_greenlets.popleft().get(block=False)
                if self._idle_event.wait(timeout=1):
                    self._idle_event.clear()
        except urwid.ExitMainLoop:
            pass


def load_consoles(filepath):
    try:
        with open(filepath, 'r') as fh:
            consoles = json.load(fh)
        return [Console.from_dict(c) for c in consoles]
    except FileNotFoundError:
        return []


def save_consoles(filepath, consoles):
    consoles = [c.to_dict() for c in consoles]
    with open(filepath, 'w') as fh:
        json.dump(consoles, fh, indent=2)


def run_tui(consoles_filepath, addr, liveid, tokens_filepath, do_refresh):
    """
    Main entrypoint for TUI

    Args:
        consoles_filepath (str): Console json filepath
        addr (str): IP address of console
        liveid (str): LiveID to connect to
        tokens_filepath (str): Tokens filepath
        do_refresh (bool): Do token refresh

    Returns:
        ExitCodes (int): Exit code
    """
    consoles = []
    if consoles_filepath:
        consoles = load_consoles(consoles_filepath)

    app = WebAPIDisplay(tokens_filepath)
    auth_success = app.run()

    if not auth_success:
        print('tui: Cannot continue without valid authentication data. bye!')
        return ExitCodes.AuthenticationError

    app = SGDisplay(consoles, app.auth_mgr)
    app.run()

    if consoles_filepath:
        save_consoles(consoles_filepath, app.consoles.consoles)

    return ExitCodes.OK
